@@ -1,5 +1,5 @@
 /*
- * Copyright 2006-2016 Adrian Thurston <thurston@colm.net>
+ * Copyright 2006-2018 Adrian Thurston <thurston@colm.net>
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to
@@ -201,7 +201,7 @@ void usage()
 void version()
 {
 	cout << "Colm version " VERSION << " " PUBDATE << endl <<
-			"Copyright (c) 2007-2016 by Adrian D. Thurston" << endl;
+			"Copyright (c) 2007-2019 by Adrian D. Thurston" << endl;
 }
 
 /* Scans a string looking for the file extension. If there is a file
@@ -415,13 +415,19 @@ void compileOutputCommand( const char *command )
 		error() << "there was a problem compiling the output" << endl;
 }
 
-void compileOutput( const char *argv0, const bool installed )
+void compileOutput( const char *argv0, const bool inSource, char *srcLocation )
 {
 	/* Find the location of the colm program that is executing. */
 	char *location = strdup( argv0 );
 	char *last;
 	int length = 1024 + strlen( intermedFn ) + strlen( binaryFn );
-	if ( installed ) {
+	if ( inSource ) {
+		last = strrchr( location, '/' );
+		assert( last != 0 );
+		last[0] = 0;
+		length += 3 * strlen( location );
+	}
+	else {
 		last = location + strlen( location ) - 1;
 		while ( true ) {
 			if ( last == location ) {
@@ -435,11 +441,6 @@ void compileOutput( const char *argv0, const bool installed )
 			}
 			last -= 1;
 		}
-	} else {
-		last = strrchr( location, '/' );
-		assert( last != 0 );
-		last[0] = 0;
-		length += 3 * strlen( location );
 	}
 	for ( ArgsVector::Iter af = additionalCodeFiles; af.lte(); af++ )
 		length += strlen( *af ) + 2;
@@ -452,20 +453,23 @@ void compileOutput( const char *argv0, const bool installed )
 		" -o %s" \
 		" %s"
 	char *command = new char[length];
-	if ( installed) {
-		sprintf( command,
-				COMPILE_COMMAND_STRING
-				" -I" PREFIX "/include"
-				" -L" PREFIX "/lib",
-				binaryFn, intermedFn );
-	}
-	else {
+	if ( inSource ) {
 		sprintf( command,
 				COMPILE_COMMAND_STRING
 				" -I%s/../aapl"
 				" -I%s/include"
-				" -L%s",
-				binaryFn, intermedFn, location, location, location );
+				" -L%s"
+				" -Wl,-rpath=%s",
+				binaryFn, intermedFn, srcLocation,
+				srcLocation, location, location );
+	}
+	else {
+		sprintf( command,
+				COMPILE_COMMAND_STRING
+				" -I" PREFIX "/include"
+				" -L" PREFIX "/lib"
+				" -Wl,-rpath," PREFIX "/lib",
+				binaryFn, intermedFn );
 	}
 #undef COMPILE_COMMAND_STRING
 	for ( ArgsVector::Iter af = additionalCodeFiles; af.lte(); af++ ) {
@@ -487,21 +491,38 @@ void compileOutput( const char *argv0, const bool installed )
 	delete[] command;
 }
 
-bool inSourceTree( const char *argv0 )
+bool inSourceTree( const char *argv0, char *&location )
 {
 	const char *lastSlash = strrchr( argv0, '/' );
 	if ( lastSlash != 0 ) {
-		int rootLen = lastSlash - argv0 + 1;
+		/* Take off the file name. */
+		int rootLen = lastSlash - argv0;
+
+		/* Create string for dir. */
 		char *mainPath = new char[rootLen + 16];
 		memcpy( mainPath, argv0, rootLen );
-		strcpy( mainPath + rootLen, "main.cc" );
+		mainPath[rootLen] = 0;
+
+		/* If built using ldconfig then there will be a .libs dir. */
+		lastSlash = strrchr( mainPath, '/' );
+		if ( lastSlash != 0 ) {
+			if ( strlen( lastSlash ) >= 6 && memcmp( lastSlash, "/.libs", 7 ) == 0 ) {
+				rootLen = lastSlash - mainPath;
+				mainPath[rootLen] = 0;
+			}
+		}
+
+		strcpy( mainPath + rootLen, "/main.cc" );
 
 		struct stat sb;
 		int res = stat( mainPath, &sb );
-		delete[] mainPath;
-
-		if ( res == 0 && S_ISREG( sb.st_mode ) )
+		if ( res == 0 && S_ISREG( sb.st_mode ) ) {
+			mainPath[rootLen] = 0;
+			location = mainPath;
 			return true;
+		}
+
+		delete[] mainPath;
 	}
 
 	return false;
@@ -744,10 +765,9 @@ int main(int argc, const char **argv)
 			delete outStream;
 
 		if ( !gblLibrary ) {
-			if ( inSourceTree( argv[0] ) )
-				compileOutput( argv[0], false );
-			else
-				compileOutput( argv[0], true );
+			char *location = 0;
+			bool inSource = inSourceTree( argv[0], location );
+			compileOutput( argv[0], inSource, location );
 		}
 
 		if ( exportHeaderFn != 0 ) {
