@@ -1,5 +1,5 @@
 /*
- * Copyright 2007-2012 Adrian Thurston <thurston@colm.net>
+ * Copyright 2007-2018 Adrian Thurston <thurston@colm.net>
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to
@@ -32,8 +32,6 @@
 #include <colm/bytecode.h>
 #include <colm/debug.h>
 
-#define BUFFER_INITIAL_SIZE 4096
-
 kid_t *alloc_attrs( program_t *prg, long length )
 {
 	kid_t *cur = 0;
@@ -255,7 +253,7 @@ static void ins_left_ignore( program_t *prg, tree_t *tree, tree_t *ignore_list )
 	/* Allocate. */
 	kid_t *kid = kid_allocate( prg );
 	kid->tree = ignore_list;
-	colm_tree_upref( ignore_list );
+	colm_tree_upref( prg, ignore_list );
 
 	/* Attach it. */
 	kid->next = tree->child;
@@ -271,7 +269,7 @@ static void ins_right_ignore( program_t *prg, tree_t *tree, tree_t *ignore_list
 	/* Insert an ignore head in the child list. */
 	kid_t *kid = kid_allocate( prg );
 	kid->tree = ignore_list;
-	colm_tree_upref( ignore_list );
+	colm_tree_upref( prg, ignore_list );
 
 	/* Attach it. */
 	if ( tree->flags & AF_LEFT_IGNORE ) {
@@ -301,7 +299,7 @@ tree_t *push_right_ignore( program_t *prg, tree_t *push_to, tree_t *right_ignore
 		 * upreffed it in insLeftIgnore. */
 		cur_ignore->tree->refs -= 1;
 		cur_ignore->tree = right_ignore;
-		colm_tree_upref( right_ignore );
+		colm_tree_upref( prg, right_ignore );
 	}
 	else {
 		/* Attach The ignore list. */
@@ -326,7 +324,7 @@ tree_t *push_left_ignore( program_t *prg, tree_t *push_to, tree_t *left_ignore )
 		 * upreffed it in insRightIgnore. */
 		cur_ignore->tree->refs -= 1;
 		cur_ignore->tree = left_ignore;
-		colm_tree_upref( left_ignore );
+		colm_tree_upref( prg, left_ignore );
 	}
 	else {
 		/* Attach the ignore list. */
@@ -379,15 +377,15 @@ tree_t *pop_right_ignore( program_t *prg, tree_t **sp, tree_t *pop_from, tree_t
 	 * right ignore. */
 	kid_t *li = tree_left_ignore_kid( prg, ri_kid->tree );
 	if ( li != 0 ) {
-		colm_tree_upref( li->tree );
+		colm_tree_upref( prg, li->tree );
 		rem_left_ignore( prg, sp, ri_kid->tree );
 		*right_ignore = ri_kid->tree;
-		colm_tree_upref( *right_ignore );
+		colm_tree_upref( prg, *right_ignore );
 		ri_kid->tree = li->tree;
 	}
 	else  {
 		*right_ignore = ri_kid->tree;
-		colm_tree_upref( *right_ignore );
+		colm_tree_upref( prg, *right_ignore );
 		rem_right_ignore( prg, sp, pop_from );
 	}
 
@@ -405,15 +403,15 @@ tree_t *pop_left_ignore( program_t *prg, tree_t **sp, tree_t *pop_from, tree_t *
 	 * left ignore. */
 	kid_t *ri = tree_right_ignore_kid( prg, li_kid->tree );
 	if ( ri != 0 ) {
-		colm_tree_upref( ri->tree );
+		colm_tree_upref( prg, ri->tree );
 		rem_right_ignore( prg, sp, li_kid->tree );
 		*left_ignore = li_kid->tree;
-		colm_tree_upref( *left_ignore );
+		colm_tree_upref( prg, *left_ignore );
 		li_kid->tree = ri->tree;
 	}
 	else {
 		*left_ignore = li_kid->tree;
-		colm_tree_upref( *left_ignore );
+		colm_tree_upref( prg, *left_ignore );
 		rem_left_ignore( prg, sp, pop_from );
 	}
 
@@ -595,7 +593,7 @@ tree_t *colm_construct_token( program_t *prg, tree_t **args, long nargs )
 		long i;
 		for ( i = 2; i < nargs; i++ ) {
 			colm_tree_set_attr( tree, i-2, args[i] );
-			colm_tree_upref( colm_get_attr( tree, i-2 ) );
+			colm_tree_upref( prg, colm_get_attr( tree, i-2 ) );
 		}
 	}
 	return tree;
@@ -706,7 +704,7 @@ tree_t *make_tree( program_t *prg, tree_t **args, long nargs )
 	for ( id = 1; id < nargs; id++ ) {
 		kid_t *kid = kid_allocate( prg );
 		kid->tree = args[id];
-		colm_tree_upref( kid->tree );
+		colm_tree_upref( prg, kid->tree );
 
 		if ( last == 0 )
 			child = kid;
@@ -760,7 +758,7 @@ kid_t *copy_kid_list( program_t *prg, kid_t *kid_list )
 		kid_t *new_ic = kid_allocate( prg );
 
 		new_ic->tree = ic->tree;
-		colm_tree_upref( new_ic->tree );
+		colm_tree_upref( prg, new_ic->tree );
 
 		/* List pointers. */
 		if ( last == 0 )
@@ -843,25 +841,12 @@ tree_t *copy_real_tree( program_t *prg, tree_t *tree, kid_t *old_next_down, kid_
 
 tree_t *colm_copy_tree( program_t *prg, tree_t *tree, kid_t *old_next_down, kid_t **new_next_down )
 {
-//	struct lang_el_info *lelInfo = prg->rtd->lelInfo;
-//	long genericId = lelInfo[tree->id].genericId;
-//	if ( genericId > 0 )
-//		assert(false);
-	if ( tree->id == LEL_ID_PTR )
-		assert(false);
-//	else if ( tree->id == LEL_ID_BOOL )
-//		assert(false);
-//	else if ( tree->id == LEL_ID_INT )
-//		assert(false);
-	else if ( tree->id == LEL_ID_STR )
-		assert(false);
-//	else if ( tree->id == LEL_ID_STREAM )
-//		assert(false);
-	else {
-		tree = copy_real_tree( prg, tree, old_next_down, new_next_down );
-	}
+	assert( tree->id != LEL_ID_PTR && tree->id != LEL_ID_STR );
+
+	tree = copy_real_tree( prg, tree, old_next_down, new_next_down );
 
 	assert( tree->refs == 0 );
+
 	return tree;
 }
 
@@ -873,7 +858,7 @@ tree_t *split_tree( program_t *prg, tree_t *tree )
 		if ( tree->refs > 1 ) {
 			kid_t *old_next_down = 0, *new_next_down = 0;
 			tree_t *new_tree = colm_copy_tree( prg, tree, old_next_down, &new_next_down );
-			colm_tree_upref( new_tree );
+			colm_tree_upref( prg, new_tree );
 
 			/* Downref the original. Don't need to consider freeing because
 			 * refs were > 1. */
@@ -892,17 +877,9 @@ tree_t *split_tree( program_t *prg, tree_t *tree )
 void tree_free_rec( program_t *prg, tree_t **sp, tree_t *tree )
 {
 	tree_t **top = vm_ptop();
-//	struct lang_el_info *lelInfo;
-//	long genericId;
 
 free_tree:
-//	lelInfo = prg->rtd->lelInfo;
-//	genericId = lelInfo[tree->id].genericId;
-//	assert( genericId == 0 );
-
 	switch ( tree->id ) {
-//	case LEL_ID_BOOL:
-//	case LEL_ID_INT:
 	case LEL_ID_PTR:
 		tree_free( prg, tree );
 		break;
@@ -941,15 +918,18 @@ free_tree:
 	}
 }
 
-void colm_tree_upref( tree_t *tree )
+void colm_tree_upref( program_t *prg, tree_t *tree )
 {
-	if ( tree != 0 )
+	if ( tree != 0 ) {
+		assert( tree->id < prg->rtd->first_struct_el_id );
 		tree->refs += 1;
+	}
 }
 
 void colm_tree_downref( program_t *prg, tree_t **sp, tree_t *tree )
 {
 	if ( tree != 0 ) {
+		assert( tree->id < prg->rtd->first_struct_el_id );
 		assert( tree->refs > 0 );
 		tree->refs -= 1;
 		if ( tree->refs == 0 )
@@ -962,11 +942,8 @@ void colm_tree_downref( program_t *prg, tree_t **sp, tree_t *tree )
 void object_free_rec( program_t *prg, tree_t **sp, tree_t *tree )
 {
 	tree_t **top = vm_ptop();
-//	struct lang_el_info *lelInfo;
-//	long genericId;
 
 free_tree:
-//	lelInfo = prg->rtd->lelInfo;
 
 	switch ( tree->id ) {
 	case LEL_ID_STR: {
@@ -975,17 +952,10 @@ free_tree:
 		tree_free( prg, tree );
 		break;
 	}
-//	case LEL_ID_BOOL:
-//	case LEL_ID_INT: {
-//		tree_free( prg, tree );
-//		break;
-//	}
 	case LEL_ID_PTR: {
 		tree_free( prg, tree );
 		break;
 	}
-//	case LEL_ID_STREAM: {
-//	}
 	default: { 
 		if ( tree->id != LEL_ID_IGNORE )
 			string_free( prg, tree->tokdata );
@@ -1335,7 +1305,7 @@ void split_ref( program_t *prg, tree_t ***psp, ref_t *from_ref )
 
 			tree_t *new_tree = colm_copy_tree( prg, ref->kid->tree, 
 					old_next_kid_down, &new_next_kid_down );
-			colm_tree_upref( new_tree );
+			colm_tree_upref( prg, new_tree );
 			
 			/* Downref the original. Don't need to consider freeing because
 			 * refs were > 1. */
@@ -1468,7 +1438,8 @@ tree_t *get_parser_mem( parser_t *parser, word_t field )
 	tree_t *result = 0;
 	switch ( field ) {
 		case 0: {
-			result = parser->result;
+			tree_t *tree = get_parsed_root( parser->pda_run, parser->pda_run->stop_target > 0 );
+			result = tree;
 			break;
 		}
 		case 1: {
@@ -1628,573 +1599,22 @@ struct colm_location *colm_find_location( program_t *prg, tree_t *tree )
 	return loc_search( prg, tree );
 }
 
-/*
- * tree_t Printing
- */
-
-void xml_escape_data( struct colm_print_args *print_args, const char *data, long len )
+head_t *tree_to_str( program_t *prg, tree_t **sp, tree_t *tree, int trim, int attrs )
 {
-	int i;
-	for ( i = 0; i < len; i++ ) {
-		if ( data[i] == '<' )
-			print_args->out( print_args, "&lt;", 4 );
-		else if ( data[i] == '>' )
-			print_args->out( print_args, "&gt;", 4 );
-		else if ( data[i] == '&' )
-			print_args->out( print_args, "&amp;", 5 );
-		else if ( (32 <= data[i] && data[i] <= 126) || 
-				data[i] == '\t' || data[i] == '\n' || data[i] == '\r' )
-		{
-			print_args->out( print_args, &data[i], 1 );
-		}
-		else {
-			char out[64];
-			sprintf( out, "&#%u;", ((unsigned)data[i]) );
-			print_args->out( print_args, out, strlen(out) );
-		}
-	}
-}
+	/* Collect the tree data. */
+	str_collect_t collect;
+	init_str_collect( &collect );
 
-void init_str_collect( StrCollect *collect )
-{
-	collect->data = (char*) malloc( BUFFER_INITIAL_SIZE );
-	collect->allocated = BUFFER_INITIAL_SIZE;
-	collect->length = 0;
-}
-
-void str_collect_destroy( StrCollect *collect )
-{
-	free( collect->data );
-}
-
-void str_collect_append( StrCollect *collect, const char *data, long len )
-{
-	long new_len = collect->length + len;
-	if ( new_len > collect->allocated ) {
-		collect->allocated = new_len * 2;
-		collect->data = (char*) realloc( collect->data, collect->allocated );
-	}
-	memcpy( collect->data + collect->length, data, len );
-	collect->length += len;
-}
-		
-void str_collect_clear( StrCollect *collect )
-{
-	collect->length = 0;
-}
-
-#define INT_SZ 32
-
-void print_str( struct colm_print_args *print_args, head_t *str )
-{
-	print_args->out( print_args, (char*)(str->data), str->length );
-}
-
-void append_collect( struct colm_print_args *args, const char *data, int length )
-{
-	str_collect_append( (StrCollect*) args->arg, data, length );
-}
-
-void append_file( struct colm_print_args *args, const char *data, int length )
-{
-	int level;
-	struct stream_impl *impl = (struct stream_impl*) args->arg;
-
-restart:
-	if ( impl->indent ) {
-		/* Consume mode. */
-		while ( length > 0 && ( *data == ' ' || *data == '\t' ) ) {
-			data += 1;
-			length -= 1;
-		}
-
-		if ( length > 0 ) {
-			/* Found some data, print the indentation and turn off indentation
-			 * mode. */
-			for ( level = 0; level < impl->level; level++ )
-				fputc( '\t', impl->file );
-
-			impl->indent = 0;
-
-			goto restart;
-		}
-	}
-	else {
-		char *nl;
-		if ( impl->level != COLM_INDENT_OFF &&
-				(nl = memchr( data, '\n', length )) )
-		{
-			/* Print up to and including the newline. */
-			int wl = nl - data + 1;
-			fwrite( data, 1, wl, impl->file );
-
-			/* Go into consume state. If we see more non-indentation chars we
-			 * will generate the appropriate indentation level. */
-			data += wl;
-			length -= wl;
-			impl->indent = 1;
-			goto restart;
-		}
-		else {
-			/* Indentation off, or no indent trigger (newline). */
-			fwrite( data, 1, length, impl->file );
-		}
-	}
-}
-
-
-tree_t *tree_trim( struct colm_program *prg, tree_t **sp, tree_t *tree )
-{
-	if ( tree == 0 )
-		return 0;
-
-	debug( prg, REALM_PARSE, "attaching left ignore\n" );
-
-	/* Make the ignore list for the left-ignore. */
-	tree_t *left_ignore = tree_allocate( prg );
-	left_ignore->id = LEL_ID_IGNORE;
-	left_ignore->flags |= AF_SUPPRESS_RIGHT;
-
-	tree = push_left_ignore( prg, tree, left_ignore );
-
-	debug( prg, REALM_PARSE, "attaching ignore right\n" );
-
-	/* Copy the ignore list first if we need to attach it as a right
-	 * ignore. */
-	tree_t *right_ignore = 0;
-	right_ignore = tree_allocate( prg );
-	right_ignore->id = LEL_ID_IGNORE;
-	right_ignore->flags |= AF_SUPPRESS_LEFT;
-
-	tree = push_right_ignore( prg, tree, right_ignore );
-
-	return tree;
-}
-
-enum ReturnType
-{
-	Done = 1,
-	CollectIgnoreLeft,
-	CollectIgnoreRight,
-	RecIgnoreList,
-	ChildPrint
-};
-
-enum VisitType
-{
-	IgnoreWrapper,
-	IgnoreData,
-	Term,
-	NonTerm
-};
-
-#define TF_TERM_SEEN 0x1
-
-void print_kid( program_t *prg, tree_t **sp, struct colm_print_args *print_args, kid_t *kid )
-{
-	enum ReturnType rt;
-	kid_t *parent = 0;
-	kid_t *leading_ignore = 0;
-	enum VisitType visit_type;
-	int flags = 0;
-
-	/* Iterate the kids passed in. We are expecting a next, which will allow us
-	 * to print the trailing ignore list. */
-	while ( kid != 0 ) {
-		vm_push_type( enum ReturnType, Done );
-		goto rec_call;
-		rec_return_top:
-		kid = kid->next;
-	}
-
-	return;
-
-rec_call:
-	if ( kid->tree == 0 )
-		goto skip_null;
-
-	/* If not currently skipping ignore data, then print it. Ignore data can
-	 * be associated with terminals and nonterminals. */
-	if ( kid->tree->flags & AF_LEFT_IGNORE ) {
-		vm_push_kid( parent );
-		vm_push_kid( kid );
-		parent = kid;
-		kid = tree_left_ignore_kid( prg, kid->tree );
-		vm_push_type( enum ReturnType, CollectIgnoreLeft );
-		goto rec_call;
-		rec_return_ign_left:
-		kid = vm_pop_kid();
-		parent = vm_pop_kid();
-	}
-
-	if ( kid->tree->id == LEL_ID_IGNORE )
-		visit_type = IgnoreWrapper;
-	else if ( parent != 0 && parent->tree->id == LEL_ID_IGNORE )
-		visit_type = IgnoreData;
-	else if ( kid->tree->id < prg->rtd->first_non_term_id )
-		visit_type = Term;
+	if ( attrs )
+		colm_print_tree_collect_a( prg, sp, &collect, tree, trim );
 	else
-		visit_type = NonTerm;
-	
-	debug( prg, REALM_PRINT, "visit type: %d\n", visit_type );
+		colm_print_tree_collect( prg, sp, &collect, tree, trim );
 
-	if ( visit_type == IgnoreData ) {
-		debug( prg, REALM_PRINT, "putting %p on ignore list\n", kid->tree );
-		kid_t *new_ignore = kid_allocate( prg );
-		new_ignore->next = leading_ignore;
-		leading_ignore = new_ignore;
-		leading_ignore->tree = kid->tree;
-		goto skip_node;
-	}
+	/* Set up the input stream. */
+	head_t *ret = string_alloc_full( prg, collect.data, collect.length );
 
-	if ( visit_type == IgnoreWrapper ) {
-		kid_t *new_ignore = kid_allocate( prg );
-		new_ignore->next = leading_ignore;
-		leading_ignore = new_ignore;
-		leading_ignore->tree = kid->tree;
-		/* Don't skip. */
-	}
+	str_collect_destroy( &collect );
 
-	/* print leading ignore? Triggered by terminals. */
-	if ( visit_type == Term ) {
-		/* Reverse the leading ignore list. */
-		if ( leading_ignore != 0 ) {
-			kid_t *ignore = 0, *last = 0;
-
-			/* Reverse the list and take the opportunity to implement the
-			 * suppress left. */
-			while ( true ) {
-				kid_t *next = leading_ignore->next;
-				leading_ignore->next = last;
-
-				if ( leading_ignore->tree->flags & AF_SUPPRESS_LEFT ) {
-					/* We are moving left. Chop off the tail. */
-					debug( prg, REALM_PRINT, "suppressing left\n" );
-					free_kid_list( prg, next );
-					break;
-				}
-
-				if ( next == 0 )
-					break;
-
-				last = leading_ignore;
-				leading_ignore = next;
-			}
-
-			/* Print the leading ignore list. Also implement the suppress right
-			 * in the process. */
-			if ( print_args->comm && (!print_args->trim ||
-					(flags & TF_TERM_SEEN && kid->tree->id > 0)) )
-			{
-				ignore = leading_ignore;
-				while ( ignore != 0 ) {
-					if ( ignore->tree->flags & AF_SUPPRESS_RIGHT )
-						break;
-
-					if ( ignore->tree->id != LEL_ID_IGNORE ) {
-						vm_push_type( enum VisitType, visit_type );
-						vm_push_kid( leading_ignore );
-						vm_push_kid( ignore );
-						vm_push_kid( parent );
-						vm_push_kid( kid );
-
-						leading_ignore = 0;
-						kid = ignore;
-						parent = 0;
-
-						debug( prg, REALM_PRINT, "rec call on %p\n", kid->tree );
-						vm_push_type( enum ReturnType, RecIgnoreList );
-						goto rec_call;
-						rec_return_il:
-
-						kid = vm_pop_kid();
-						parent = vm_pop_kid();
-						ignore = vm_pop_kid();
-						leading_ignore = vm_pop_kid();
-						visit_type = vm_pop_type(enum VisitType);
-					}
-
-					ignore = ignore->next;
-				}
-			}
-
-			/* Free the leading ignore list. */
-			free_kid_list( prg, leading_ignore );
-			leading_ignore = 0;
-		}
-	}
-
-	if ( visit_type == Term || visit_type == NonTerm ) {
-		/* Open the tree. */
-		print_args->open_tree( prg, sp, print_args, parent, kid );
-	}
-
-	if ( visit_type == Term )
-		flags |= TF_TERM_SEEN;
-
-	if ( visit_type == Term || visit_type == IgnoreData ) {
-		/* Print contents. */
-		if ( kid->tree->id < prg->rtd->first_non_term_id ) {
-			debug( prg, REALM_PRINT, "printing terminal %p\n", kid->tree );
-			if ( kid->tree->id != 0 )
-				print_args->print_term( prg, sp, print_args, kid );
-		}
-	}
-
-	/* Print children. */
-	kid_t *child = print_args->attr ? 
-		tree_attr( prg, kid->tree ) : 
-		tree_child( prg, kid->tree );
-
-	if ( child != 0 ) {
-		vm_push_type( enum VisitType, visit_type );
-		vm_push_kid( parent );
-		vm_push_kid( kid );
-		parent = kid;
-		kid = child;
-		while ( kid != 0 ) {
-			vm_push_type( enum ReturnType, ChildPrint );
-			goto rec_call;
-			rec_return:
-			kid = kid->next;
-		}
-		kid = vm_pop_kid();
-		parent = vm_pop_kid();
-		visit_type = vm_pop_type(enum VisitType);
-	}
-
-	if ( visit_type == Term || visit_type == NonTerm ) {
-		/* close the tree. */
-		print_args->close_tree( prg, sp, print_args, parent, kid );
-	}
-
-skip_node:
-
-	/* If not currently skipping ignore data, then print it. Ignore data can
-	 * be associated with terminals and nonterminals. */
-	if ( kid->tree->flags & AF_RIGHT_IGNORE ) {
-		debug( prg, REALM_PRINT, "right ignore\n" );
-		vm_push_kid( parent );
-		vm_push_kid( kid );
-		parent = kid;
-		kid = tree_right_ignore_kid( prg, kid->tree );
-		vm_push_type( enum ReturnType, CollectIgnoreRight );
-		goto rec_call;
-		rec_return_ign_right:
-		kid = vm_pop_kid();
-		parent = vm_pop_kid();
-	}
-
-/* For skiping over content on null. */
-skip_null:
-
-	rt = vm_pop_type(enum ReturnType);
-	switch ( rt ) {
-		case Done:
-			debug( prg, REALM_PRINT, "return: done\n" );
-			goto rec_return_top;
-			break;
-		case CollectIgnoreLeft:
-			debug( prg, REALM_PRINT, "return: ignore left\n" );
-			goto rec_return_ign_left;
-		case CollectIgnoreRight:
-			debug( prg, REALM_PRINT, "return: ignore right\n" );
-			goto rec_return_ign_right;
-		case RecIgnoreList:
-			debug( prg, REALM_PRINT, "return: ignore list\n" );
-			goto rec_return_il;
-		case ChildPrint:
-			debug( prg, REALM_PRINT, "return: child print\n" );
-			goto rec_return;
-	}
+	return ret;
 }
 
-void colm_print_tree_args( program_t *prg, tree_t **sp,
-		struct colm_print_args *print_args, tree_t *tree )
-{
-	if ( tree == 0 )
-		print_args->out( print_args, "NIL", 3 );
-	else {
-		/* This term tree allows us to print trailing ignores. */
-		tree_t term_tree;
-		memset( &term_tree, 0, sizeof(term_tree) );
-
-		kid_t kid, term;
-		term.tree = &term_tree;
-		term.next = 0;
-		term.flags = 0;
-
-		kid.tree = tree;
-		kid.next = &term;
-		kid.flags = 0;
-
-		print_kid( prg, sp, print_args, &kid );
-	}
-}
-
-void colm_print_term_tree( program_t *prg, tree_t **sp,
-		struct colm_print_args *print_args, kid_t *kid )
-{
-	debug( prg, REALM_PRINT, "printing term %p\n", kid->tree );
-
-	if ( kid->tree->id == LEL_ID_PTR ) {
-		char buf[INT_SZ];
-		print_args->out( print_args, "#<", 2 );
-		sprintf( buf, "%lx", ((pointer_t*)kid->tree)->value );
-		print_args->out( print_args, buf, strlen(buf) );
-		print_args->out( print_args, ">", 1 );
-	}
-	else if ( kid->tree->id == LEL_ID_STR ) {
-		print_str( print_args, ((str_t*)kid->tree)->value );
-	}
-//	else if ( kid->tree->id == LEL_ID_STREAM ) {
-//		char buf[INT_SZ];
-//		printArgs->out( printArgs, "#", 1 );
-//		sprintf( buf, "%p", (void*) ((stream_t*)kid->tree)->in->file );
-//		printArgs->out( printArgs, buf, strlen(buf) );
-//	}
-	else if ( kid->tree->tokdata != 0 && 
-			string_length( kid->tree->tokdata ) > 0 )
-	{
-		print_args->out( print_args, string_data( kid->tree->tokdata ), 
-				string_length( kid->tree->tokdata ) );
-	}
-
-	struct lang_el_info *lel_info = prg->rtd->lel_info;
-	struct stream_impl *impl = (struct stream_impl*) print_args->arg;
-
-	if ( strcmp( lel_info[kid->tree->id].name, "_IN_" ) == 0 ) {
-		if ( impl->level == COLM_INDENT_OFF ) {
-			impl->level = 1;
-			impl->indent = 1;
-		}
-		else {
-			impl->level += 1;
-		}
-	}
-
-	if ( strcmp( lel_info[kid->tree->id].name, "_EX_" ) == 0 )
-		impl->level -= 1;
-}
-
-
-void colm_print_null( program_t *prg, tree_t **sp,
-		struct colm_print_args *args, kid_t *parent, kid_t *kid )
-{
-}
-
-void open_tree_xml( program_t *prg, tree_t **sp, struct colm_print_args *args,
-		kid_t *parent, kid_t *kid )
-{
-	/* Skip the terminal that is for forcing trailing ignores out. */
-	if ( kid->tree->id == 0 )
-		return;
-
-	struct lang_el_info *lel_info = prg->rtd->lel_info;
-
-	/* List flattening: skip the repeats and lists that are a continuation of
-	 * the list. */
-	if ( parent != 0 && parent->tree->id == kid->tree->id && kid->next == 0 &&
-			( lel_info[parent->tree->id].repeat || lel_info[parent->tree->id].list ) )
-	{
-		return;
-	}
-
-	const char *name = lel_info[kid->tree->id].xml_tag;
-	args->out( args, "<", 1 );
-	args->out( args, name, strlen( name ) );
-	args->out( args, ">", 1 );
-}
-
-void print_term_xml( program_t *prg, tree_t **sp,
-	struct colm_print_args *print_args, kid_t *kid )
-{
-	//kid_t *child;
-
-	/*child = */ tree_child( prg, kid->tree );
-	if ( kid->tree->id == LEL_ID_PTR ) {
-		char ptr[INT_SZ];
-		sprintf( ptr, "%lx", ((pointer_t*)kid->tree)->value );
-		print_args->out( print_args, ptr, strlen(ptr) );
-	}
-	else if ( kid->tree->id == LEL_ID_STR ) {
-		head_t *head = (head_t*) ((str_t*)kid->tree)->value;
-
-		xml_escape_data( print_args, (char*)(head->data), head->length );
-	}
-	else if ( 0 < kid->tree->id && kid->tree->id < prg->rtd->first_non_term_id &&
-			kid->tree->id != LEL_ID_IGNORE &&
-			kid->tree->tokdata != 0 && 
-			string_length( kid->tree->tokdata ) > 0 )
-	{
-		xml_escape_data( print_args, string_data( kid->tree->tokdata ), 
-				string_length( kid->tree->tokdata ) );
-	}
-}
-
-
-void close_tree_xml( program_t *prg, tree_t **sp,
-		struct colm_print_args *args, kid_t *parent, kid_t *kid )
-{
-	/* Skip the terminal that is for forcing trailing ignores out. */
-	if ( kid->tree->id == 0 )
-		return;
-
-	struct lang_el_info *lel_info = prg->rtd->lel_info;
-
-	/* List flattening: skip the repeats and lists that are a continuation of
-	 * the list. */
-	if ( parent != 0 && parent->tree->id == kid->tree->id && kid->next == 0 &&
-			( lel_info[parent->tree->id].repeat || lel_info[parent->tree->id].list ) )
-	{
-		return;
-	}
-
-	const char *name = lel_info[kid->tree->id].xml_tag;
-	args->out( args, "</", 2 );
-	args->out( args, name, strlen( name ) );
-	args->out( args, ">", 1 );
-}
-
-void print_tree_collect( program_t *prg, tree_t **sp,
-		StrCollect *collect, tree_t *tree, int trim )
-{
-	struct colm_print_args print_args = {
-			collect, true, false, trim, &append_collect, 
-			&colm_print_null, &colm_print_term_tree, &colm_print_null
-	};
-
-	colm_print_tree_args( prg, sp, &print_args, tree );
-}
-
-void print_tree_collect_a( program_t *prg, tree_t **sp,
-		StrCollect *collect, tree_t *tree, int trim )
-{
-	struct colm_print_args print_args = {
-			collect, true, true, trim, &append_collect, 
-			&colm_print_null, &colm_print_term_tree, &colm_print_null
-	};
-
-	colm_print_tree_args( prg, sp, &print_args, tree );
-}
-
-void print_tree_file( program_t *prg, tree_t **sp, struct stream_impl *impl,
-		tree_t *tree, int trim )
-{
-	struct colm_print_args print_args = {
-			impl, true, false, trim, &append_file, 
-			&colm_print_null, &colm_print_term_tree, &colm_print_null
-	};
-
-	colm_print_tree_args( prg, sp, &print_args, tree );
-}
-
-void print_xml_stdout( program_t *prg, tree_t **sp,
-		struct stream_impl *impl, tree_t *tree,
-		int comm_attr, int trim )
-{
-	struct colm_print_args print_args = {
-			impl, comm_attr, comm_attr, trim, &append_file, 
-			&open_tree_xml, &print_term_xml, &close_tree_xml };
-	colm_print_tree_args( prg, sp, &print_args, tree );
-}
